/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
 */

package ch.epfl.lia.frodo.algorithms;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;

import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.ProblemInterface;

/** A solver for a general problem
 * @author Thomas Leaute
 * @param <P> type used for the problem
 * @param <V> type used for variable values
 * @param <U> type used for utility values
 * @param <S> type used for the solution
 */
public abstract class AbstractSolver < P extends ProblemInterface<V, U>, V extends Addable<V>, U extends Addable<U>, S extends Solution<V,U> > {

	/** Description of the agent to be used */
	protected Document agentDesc;
	
	/** The agent factory */
	protected AgentFactory<V> factory;

	/** The class of the parser to be used */
	protected Class< ? extends XCSPparser<V, U> > parserClass; /// @bug MAS solvers don't use the XCSPparser
	
	/** The list of modules that record statistics about problem solving */
	protected List<? extends StatsReporter> solGatherers;
	
	/** The problem */
	protected P problem;
	
	/** Whether to use TCP pipes or shared memory pipes */
	protected final boolean useTCP;
	
	/**
	 * Dummy constructor
	 */
	protected AbstractSolver() {
		this.useTCP = false;
	}
	
	/** Constructor from an agent configuration file
	 * @param agentDescFile 	the agent configuration file
	 */
	protected AbstractSolver (String agentDescFile) {
		this (agentDescFile, false);
	}
	
	/** Constructor from an agent configuration file
	 * @param agentDescFile 	the agent configuration file
	 * @param useTCP 			Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractSolver (String agentDescFile, boolean useTCP) {
		try {
			this.agentDesc = XCSPparser.parse(AgentFactory.class.getResourceAsStream(agentDescFile), false);
		} catch (JDOMException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.setParserClass();
		
		this.useTCP = useTCP;
		if (this.useTCP) // disable simulated time
			this.agentDesc.getRootElement().setAttribute("measureTime", "false");
	}
	
	/** Constructor
	 * @param agentDesc 	a JDOM Document for the agent description
	 */
	protected AbstractSolver (Document agentDesc) {
		this(agentDesc, false);
	}
	
	/** Constructor
	 * @param agentDesc 	a JDOM Document for the agent description
	 * @param useTCP 		Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractSolver (Document agentDesc, boolean useTCP) {
		this.agentDesc = agentDesc;
		this.setParserClass();
		
		this.useTCP = useTCP;
		if (this.useTCP) // disable simulated time
			this.agentDesc.getRootElement().setAttribute("measureTime", "false");
	}
	
	/** Sets the parser class as specified in the agent configuration file */
	@SuppressWarnings("unchecked")
	private void setParserClass () {
		Element parserElmt = agentDesc.getRootElement().getChild("parser");
		if (parserElmt != null) {
			try {
				this.parserClass = (Class<? extends XCSPparser<V, U>>) Class.forName(parserElmt.getAttributeValue("parserClass"));
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			}
		} else {
			this.parserClass = (Class<? extends XCSPparser<V, U>>) XCSPparser.class;
			parserElmt = new Element ("parser");
			agentDesc.getRootElement().addContent(parserElmt);
			parserElmt.setAttribute("className", this.parserClass.getName());
		}
	}
	
	/** Constructor
	 * @param agentDesc 	The agent description
	 * @param parserClass 	The class of the parser to be used
	 */
	protected AbstractSolver (Document agentDesc, Class< ? extends XCSPparser<V, U> > parserClass) {
		this(agentDesc, parserClass, false);
	}
	
	/** Constructor
	 * @param agentDesc 	The agent description
	 * @param parserClass 	The class of the parser to be used
	 * @param useTCP 		Whether to use TCP pipes or shared memory pipes
	 * @warning Using TCP pipes automatically disables simulated time. 
	 */
	protected AbstractSolver (Document agentDesc, Class< ? extends XCSPparser<V, U> > parserClass, boolean useTCP) {
		this.agentDesc = agentDesc;
		this.parserClass = parserClass;
		
		this.useTCP = useTCP;
		if (this.useTCP) // disable simulated time
			this.agentDesc.getRootElement().setAttribute("measureTime", "false");
	}
	
	/** Sets the class for variable values
	 * @param domClass 	the class for variable values
	 */
	public void setDomClass (Class<V> domClass) {
		this.agentDesc.getRootElement().getChild("parser").setAttribute("domClass", domClass.getName());
	}
	
	/** Sets the class for utility values
	 * @param utilClass 	the class for utility values
	 */
	public void setUtilClass(Class<U> utilClass) {
		this.agentDesc.getRootElement().getChild("parser").setAttribute("utilClass", utilClass.getName());
	}

	/** @return The list of modules that record statistics about problem solving */
	protected abstract List<? extends StatsReporter> getSolGatherers ();
	
	/** @return the solution as seen by the solution gatherers */
	protected abstract S buildSolution ();
	
	/** Solves the input problem
	 * @param problem 		the problem
	 * @return 				an optimal solution
	 */
	public S solve (Document problem) {
		return solve(problem, true);
	}
	
	/** Solves the input problem
	 * @param problem 		the problem
	 * @return 				an optimal solution
	 */
	public S solve (P problem) {
		return solve(problem, true);
	}
	
	/** Solves the input problem
	 * @param problem 		the problem
	 * @param timeout 		timeout in ms, if \c null, no timeout is used
	 * @return 				an optimal solution
	 */
	public S solve (Document problem, Long timeout) {
		return solve(problem, true, timeout);
	}
	
	/** Solves the input problem
	 * @param problem 		the problem
	 * @param timeout 		timeout in ms, if \c null, no timeout is used
	 * @return 				an optimal solution
	 */
	public S solve (P problem, Long timeout) {
		return solve(problem, true, timeout);
	}
	
	/** Solves the input problem
	 * @param problem 			the problem
	 * @param cleanAfterwards 	if \c true, cleans all the agents and the queue when they're done
	 * @return 					an optimal solution
	 */
	public S solve (Document problem, boolean cleanAfterwards) {
		return solve(problem, cleanAfterwards, null);
	}
	
	/** Solves the input problem
	 * @param problem 			the problem
	 * @param cleanAfterwards 	if \c true, cleans all the agents and the queue when they're done
	 * @return 					an optimal solution
	 */
	public S solve (P problem, boolean cleanAfterwards) {
		return solve(problem, cleanAfterwards, null);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @return 						an optimal solution
	 */
	public S solve (Document problem, int nbrElectionRounds) {
		return solve (problem, nbrElectionRounds, false);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @return 						an optimal solution
	 */
	public S solve (P problem, int nbrElectionRounds) {
		return solve (problem, nbrElectionRounds, false);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @return 						an optimal solution
	 */
	public S solve (Document problem, int nbrElectionRounds, boolean measureMsgs) {
		return solve (problem, nbrElectionRounds, measureMsgs, null);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @return 						an optimal solution
	 */
	public S solve (P problem, int nbrElectionRounds, boolean measureMsgs) {
		return solve (problem, nbrElectionRounds, measureMsgs, null);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @return 						an optimal solution
	 */
	public S solve (Document problem, int nbrElectionRounds, boolean measureMsgs, Long timeout) {
		return solve(problem, nbrElectionRounds, measureMsgs, timeout, true);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @return 						an optimal solution
	 */
	public S solve (P problem, int nbrElectionRounds, boolean measureMsgs, Long timeout) {
		return solve(problem, nbrElectionRounds, measureMsgs, timeout, true);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @return 						an optimal solution
	 */
	public S solve (Document problem, int nbrElectionRounds, Long timeout) {
		return solve(problem, nbrElectionRounds, false, timeout, true);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @return 						an optimal solution
	 */
	public S solve (P problem, int nbrElectionRounds, Long timeout) {
		return solve(problem, nbrElectionRounds, false, timeout, true);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @param cleanAfterwards 		if \c true, cleans all the agents and the queue when they're done
	 * @return 						an optimal solution
	 */
	public S solve (Document problem, int nbrElectionRounds, boolean measureMsgs, Long timeout, boolean cleanAfterwards) {
		
		agentDesc.getRootElement().setAttribute("measureMsgs", Boolean.toString(measureMsgs));
		return solve(problem, cleanAfterwards, timeout);
	}
	
	/** Solves the input problem
	 * @param problem 				the problem
	 * @param nbrElectionRounds 	the number of rounds of VariableElection (must be greater than the diameter of the constraint graph)
	 * @param measureMsgs 			whether message sizes should be measured
	 * @param timeout 				timeout in ms, if \c null, no timeout is used
	 * @param cleanAfterwards 		if \c true, cleans all the agents and the queue when they're done
	 * @return 						an optimal solution
	 */
	public S solve (P problem, int nbrElectionRounds, boolean measureMsgs, Long timeout, boolean cleanAfterwards) {
		
		agentDesc.getRootElement().setAttribute("measureMsgs", Boolean.toString(measureMsgs));
		return solve(problem, cleanAfterwards, timeout);
	}
	
	/** Solves the input problem
	 * @param problem 			the problem
	 * @param cleanAfterwards 	if \c true, cleans all the agents and the queue when they're done
	 * @param timeout 			timeout in ms, if \c null, no timeout is used
	 * @return 					an optimal solution
	 */
	@SuppressWarnings("unchecked")
	public S solve (Document problem, boolean cleanAfterwards, Long timeout) {
		
		// Instantiate the parser
		Element parserElmt = this.agentDesc.getRootElement().getChild("parser");
		parserElmt.setAttribute("displayGraph", "false");
		try {
			return solve ((P) this.parserClass.getConstructor(Document.class, Element.class).newInstance(problem, parserElmt), cleanAfterwards, timeout);
		} catch (InstantiationException e) {
			System.err.println("The parser class " + this.parserClass + " is abstract");
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			System.err.println("The constructor for " + this.parserClass + " is inaccessible");
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			System.err.println("The constructor for " + this.parserClass + " threw an exception");
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			System.err.println("The parser class " + this.parserClass + " does not have a constructor that takes in a Document");
			e.printStackTrace();
		}
		
		return null;
	}
	
	/** Solves the input problem
	 * @param problem 			the problem
	 * @param cleanAfterwards 	if \c true, cleans all the agents and the queue when they're done
	 * @param timeout 			timeout in ms, if \c null, no timeout is used
	 * @return 					an optimal solution
	 */
	public S solve (P problem, boolean cleanAfterwards, Long timeout) {
		
		this.problem = problem;
		
		// Check whether it is necessary to instantiate the agent factory
		if (this.factory == null) {

			// Instantiate the modules that listen for the solution
			this.solGatherers = this.getSolGatherers();

			// Solve the problem
			this.factory = new AgentFactory<V> (problem, agentDesc, solGatherers, timeout, this.useTCP);
			
		} else { // we already have an agent factory; just restart it
			
			for (StatsReporter module : this.solGatherers) 
				module.reset();
			this.factory.restart(problem);
		}
		
		if(factory.timedOut() || factory.outOfMemory()) {
			if (cleanAfterwards)
				this.clear();
			return null;
		}
		
		S solution = this.buildSolution();
		
		if (cleanAfterwards) 
			this.clear();

		return solution;
	}

	/** Clears the parser */
	protected void clear() {
		for (StatsReporter module : this.solGatherers) 
			module.reset();
		this.solGatherers = null;
		factory.end();
		factory = null;
	}
}
