/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.dpop.param.test;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import junit.extensions.RepeatedTest;
import junit.framework.TestSuite;
import ch.epfl.lia.frodo.algorithms.RandGraphFactory;
import ch.epfl.lia.frodo.algorithms.dpop.param.ParamUTIL;
import ch.epfl.lia.frodo.algorithms.dpop.param.ParamVALUE;
import ch.epfl.lia.frodo.algorithms.dpop.param.ParamUTIL.OptUtilMessage;
import ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest;
import ch.epfl.lia.frodo.communication.Message;
import ch.epfl.lia.frodo.communication.Queue;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.BasicUtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.BasicHypercube;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.Hypercube;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.ScalarHypercube;

/** JUnit test for the class ParamVALUE
 * @author Thomas Leaute
 */
public class ParamVALUEtest extends VALUEpropagationTest {

	/** Maximum number of variables in the random graph 
	 * @note Must be at least 2. 
	 */
	private final int maxNbrVars = 5;
	
	/** Maximum number of edges in the random graph */
	private final int maxNbrEdges = 25;

	/** Maximum number of agents */
	private final int maxNbrAgents = 5;

	/** The optimal assignment to variables, conditioned on the values of the parameters */
	private Map< String[], BasicUtilitySolutionSpace<AddableInteger, ArrayList<AddableInteger> > > optAssignments;

	/** Constructor that instantiates a test only for the input method
	 * @param method test method
	 */
	public ParamVALUEtest(String method) {
		super (method);
	}

	/** @return the test suite for this test */
	static public TestSuite suite () {
		TestSuite testSuite = new TestSuite ("Tests for ParamVALUE");
		
		TestSuite testTmp = new TestSuite ("Tests for the VALUE propagation protocol using shared memory pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomSharedMemoryInt"), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol with XML support using shared memory pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomSharedMemoryIntXML"), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol using TCP pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomTCPint"), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol with XML support using TCP pipes and integer utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomTCPintXML"), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol using shared memory pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomSharedMemoryReal"), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol with XML support using shared memory pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomSharedMemoryRealXML"), 200));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol using TCP pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomTCPreal"), 100));
		testSuite.addTest(testTmp);
		
		testTmp = new TestSuite ("Tests for the VALUE propagation protocol with XML support using TCP pipes and real utilities");
		testTmp.addTest(new RepeatedTest (new ParamVALUEtest ("testRandomTCPrealXML"), 100));
		testSuite.addTest(testTmp);
		
		return testSuite;
	}

	/** @see VALUEpropagationTest#setUp() */
	@Override
	protected void setUp() {
		graph = RandGraphFactory.getRandGraph(maxNbrVars, maxNbrEdges, maxNbrAgents);
		dfs = ParamUTILtest.computeDFS(graph);
		maximize = (Math.random() < 0.5);
		optAssignments = new HashMap< String[], BasicUtilitySolutionSpace<AddableInteger, ArrayList<AddableInteger> > > (graph.nodes.size());
	}
	
	/** @see VALUEpropagationTest#tearDown() */
	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
		this.optAssignments = null;
	}

	/** @see VALUEpropagationTest#newListenerInstance(boolean, boolean, java.lang.Class) */
	@Override
	protected < U extends Addable<U> > VALUEpropagationTest.Listener<U> newListenerInstance(boolean useTCP, boolean useXML, Class<U> utilClass) 
	throws Exception {
		return new Listener<U> (useTCP, useXML, utilClass);
	}

	/** The listener that checks the messages sent by the ParamUTIL and ParamVALUE listeners
	 * @param <U> 	the type used for utility values
	 */
	protected class Listener < U extends Addable<U> > extends VALUEpropagationTest.Listener<U> {
		
		/** Optimal utility of each connected component (identified by its root), conditioned on the values of the parameters (if any) */
		private HashMap< String, UtilitySolutionSpace<AddableInteger, U> > optUtils = new HashMap< String, UtilitySolutionSpace<AddableInteger, U> > ();
		
		/** Used to synchronize access to optUtils, which can be \c null */
		private final Object optUtils_lock = new Object ();
		
		/** Constructor that tests the UTIL and VALUE propagation protocols on a random DFS 
		 * @param useTCP 						\c true whether TCP pipes should be used instead of QueueIOPipes
		 * @param useXML 						whether we should use the constructor that takes in XML elements, or the manual constructor
		 * @param utilClass 					the class to be used for utility values
		 * @throws IOException 					if the method fails to create pipes
		 * @throws NoSuchMethodException 		if the ParamVALUE class does not have a public constructor that takes in a ProblemInterface and a JDOM Element
		 * @throws InvocationTargetException 	if the ParamVALUE constructor throws an exception
		 * @throws IllegalAccessException 		if the ParamVALUE class does not have a public constructor that takes in a ProblemInterface and a JDOM Element
		 * @throws InstantiationException 		would be thrown if ParamVALUE were abstract
		 * @throws IllegalArgumentException 	if the ParamVALUE constructor does not take the proper arguments
		 */
		public Listener (boolean useTCP, boolean useXML, Class<U> utilClass) 
		throws IOException, NoSuchMethodException, IllegalArgumentException, 
		InstantiationException, IllegalAccessException, InvocationTargetException {
			super (useTCP, useXML, utilClass, ParamUTIL.class, ParamVALUE.class, true);
		}
		
		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest.Listener#checkOutput() */
		@Override
		protected void checkOutput() {
			
			// Check that the assignments found by the protocol are indeed optimal
			HashMap< String, BasicHypercube<AddableInteger, U> > optUtils2 = new HashMap< String, BasicHypercube<AddableInteger, U> > ();
			for (String var : graph.nodes) {
				
				// Check if the variable is a root
				if (dfs.get(var).getParent() == null) {
					
					// Simulate the UTIL propagation with slices instead of projections 
					// to compute the utility value for this root's DFS tree corresponding to 
					// the optimal assignments computed by the VALUE propagation protocol
					optUtils2.put(var, simulateUTILslice (var));
				}
			}
			
			// Check that the optimal utility values are the same
			assertEquals (optUtils2.size(), optUtils.size());
			for (Map.Entry< String, UtilitySolutionSpace<AddableInteger, U> > entry : optUtils.entrySet()) 
				assertTrue (optUtils2.get(entry.getKey()) + " != " + entry.getValue(), optUtils2.get(entry.getKey()).equivalent(entry.getValue()));

		}
		
		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest.Listener#getMsgTypes() */
		@Override
		public Collection<String> getMsgTypes() {
			ArrayList<String> types = new ArrayList<String> (2);
			types.add(ParamVALUE.PARAM_OUTPUT_MSG_TYPE);
			types.add(ParamUTIL.OPT_PARAM_UTIL_MSG_TYPE);
			return types;
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest.Listener#notifyIn(Message) */
		@Override
		@SuppressWarnings("unchecked")
		public void notifyIn(Message msg) {

			String type = msg.getType();
			
			if (type.equals(ParamUTIL.OPT_PARAM_UTIL_MSG_TYPE)) { // message sent by a root containing the optimal utility value
				OptUtilMessage<AddableInteger, U> msgCast = (OptUtilMessage<AddableInteger, U>) msg;
				synchronized (optUtils_lock) {
					optUtils.put(msgCast.getRoot(), msgCast.getUtility());
				}
			}
			
			else if (type.equals(ParamVALUE.PARAM_OUTPUT_MSG_TYPE)) { // optimal assignment to a variable
				ParamVALUE.AssignmentMessage<AddableInteger> msgCast = (ParamVALUE.AssignmentMessage<AddableInteger>) msg;
				synchronized (optAssignments) {
					optAssignments.put(msgCast.getVariables(), msgCast.getValues());
				}
			}
			
			finished_lock.lock();
			if (--nbrMsgsRemaining <= 0) 
				finished.signal();
			finished_lock.unlock();
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest.Listener#getStatsFromQueue(ch.epfl.lia.frodo.communication.Queue) */
		@Override
		public void getStatsFromQueue(Queue queue) {
			queue.addIncomingMessagePolicy(ParamVALUE.PARAM_OUTPUT_MSG_TYPE, this);
		}

		/** @see ch.epfl.lia.frodo.algorithms.dpop.test.VALUEpropagationTest.Listener#simulateUTILslice(java.lang.String) */
		@Override
		protected Hypercube<AddableInteger, U> simulateUTILslice (String var) {
			
			List<String> children = dfs.get(var).getChildren();
			
			// Look up the optimal assignments to var
			String[] vars = null;
			BasicUtilitySolutionSpace<AddableInteger, ArrayList<AddableInteger> > assignments = null;
			for (Map.Entry< String[], BasicUtilitySolutionSpace<AddableInteger, ArrayList<AddableInteger> > > entry : optAssignments.entrySet()) {
				vars = entry.getKey();
				if (vars[0].equals(var)) {
					assignments = entry.getValue();
					break;
				}
			}
			
			if (children.size() == 0) { // leaf variable
				
				// Check if the variable is unconstrained
				UtilitySolutionSpace<AddableInteger, U> space = hypercubes.get(var);
				if (space == null) 
					return new ScalarHypercube<AddableInteger, U> (parser.getZeroUtility(), (parser.maximize() ? parser.getMinInfUtility() : parser.getPlusInfUtility()), new AddableInteger [0].getClass());
				
				// Slice variable out of its private hypercube and return result
				return (Hypercube<AddableInteger, U>) space.compose(vars, assignments);
				
			} else { // non-leaf variable

				// Compute the hypercube received from each child and process it to compute the join
				UtilitySolutionSpace<AddableInteger, U> join = hypercubes.get(var);
				for ( String child : children ) {
					if (join == null) {
						join = simulateUTILslice(child);
					} else 
						join = join.join(simulateUTILslice(child));
				}
				
				// Slice out the variable and return the result
				return (Hypercube<AddableInteger, U>) join.compose(vars, assignments);
			}
		}
		
	}
}
